@@ -12,6 +12,13 @@ import virtualenv
 from process_tests import TestProcess as _TestProcess
 from process_tests import dump_on_error
 from process_tests import wait_for_strings
+from six import exec_
+from fields import Namespace
+
+try:
+    from StringIO import StringIO
+except ImportError:
+    from io import StringIO
 
 import pytest_cov.plugin
 
@@ -135,27 +142,12 @@ def test_foo(cov):
     assert cov is None
 '''
 
-MULTIPROCESSING_SCRIPT = '''
-import multiprocessing
-
-def target_fn():
-    a = True
-    return a
-
-def test_run_target():
-    p = multiprocessing.Process(target=target_fn)
-    p.start()
-    p.join()
-'''
-
 SCRIPT_FAIL = '''
 def test_fail():
     assert False
 
 '''
 
-SCRIPT_RESULT = '9 * 89%'
-SCRIPT2_RESULT = '3 * 100%'
 CHILD_SCRIPT_RESULT = '[56] * 100%'
 PARENT_SCRIPT_RESULT = '8 * 100%'
 DEST_DIR = 'cov_dest'
@@ -164,17 +156,37 @@ REPORT_NAME = 'cov.xml'
 xdist = pytest.mark.parametrize('opts', ['', '-n 1'], ids=['nodist', 'xdist'])
 
 
-def test_central(testdir):
-    script = testdir.makepyfile(SCRIPT)
+@pytest.fixture(params=[
+    ('branch=true', '--cov-branch', '9 * 85%', '3 * 100%'),
+    ('branch=true', '',             '9 * 85%', '3 * 100%'),
+    ('',            '--cov-branch', '9 * 85%', '3 * 100%'),
+    ('',            '',             '9 * 89%', '3 * 100%'),
+], ids=['branch2x', 'branch1c', 'branch1a', 'nobranch'])
+def prop(request):
+    return Namespace(
+        code=SCRIPT,
+        code2=SCRIPT2,
+        conf=request.param[0],
+        fullconf='[run]\n%s\n' % request.param[0],
+        args=request.param[1].split(),
+        result=request.param[2],
+        result2=request.param[3],
+    )
+
+
+def test_central(testdir, prop):
+    script = testdir.makepyfile(prop.code)
+    testdir.tmpdir.join('.coveragerc').write(prop.fullconf)
 
     result = testdir.runpytest('-v',
                                '--cov=%s' % script.dirpath(),
                                '--cov-report=term-missing',
-                               script)
+                               script,
+                               *prop.args)
 
     result.stdout.fnmatch_lines([
         '*- coverage: platform *, python * -*',
-        'test_central* %s *' % SCRIPT_RESULT,
+        'test_central* %s *' % prop.result,
         '*10 passed*'
     ])
     assert result.ret == 0
@@ -215,7 +227,7 @@ def test_annotate_output_dir(testdir):
     assert result.ret == 0
 
 
-def test_html_output_dir(testdir):
+def test_html_output_dir(testdir,prop):
     script = testdir.makepyfile(SCRIPT)
 
     result = testdir.runpytest('-v',
@@ -296,6 +308,9 @@ def test_cov_min_100(testdir):
                                script)
 
     assert result.ret != 0
+    result.stdout.fnmatch_lines([
+        'FAIL Required test coverage of 100% not reached. Total coverage: *%'
+    ])
 
 
 def test_cov_min_50(testdir):
@@ -308,6 +323,9 @@ def test_cov_min_50(testdir):
                                script)
 
     assert result.ret == 0
+    result.stdout.fnmatch_lines([
+        'Required test coverage of 50% reached. Total coverage: *%'
+    ])
 
 
 def test_cov_min_no_report(testdir):
@@ -320,19 +338,22 @@ def test_cov_min_no_report(testdir):
                                script)
 
     assert result.ret == 0
+    result.stdout.fnmatch_lines([
+        'Required test coverage of 50% reached. Total coverage: *%'
+    ])
 
 
-def test_central_nonspecific(testdir):
-    script = testdir.makepyfile(SCRIPT)
-
+def test_central_nonspecific(testdir, prop):
+    script = testdir.makepyfile(prop.code)
+    testdir.tmpdir.join('.coveragerc').write(prop.fullconf)
     result = testdir.runpytest('-v',
                                '--cov',
                                '--cov-report=term-missing',
-                               script)
+                               script, *prop.args)
 
     result.stdout.fnmatch_lines([
         '*- coverage: platform *, python * -*',
-        'test_central_nonspecific* %s *' % SCRIPT_RESULT,
+        'test_central_nonspecific* %s *' % prop.result,
         '*10 passed*'
     ])
 
@@ -358,18 +379,18 @@ fail_under = 100
     assert result.ret != 0
 
 
-def test_central_coveragerc(testdir):
-    script = testdir.makepyfile(SCRIPT)
-    testdir.tmpdir.join('.coveragerc').write(COVERAGERC_SOURCE)
+def test_central_coveragerc(testdir, prop):
+    script = testdir.makepyfile(prop.code)
+    testdir.tmpdir.join('.coveragerc').write(COVERAGERC_SOURCE + prop.conf)
 
     result = testdir.runpytest('-v',
                                '--cov',
                                '--cov-report=term-missing',
-                               script)
+                               script, *prop.args)
 
     result.stdout.fnmatch_lines([
         '*- coverage: platform *, python * -*',
-        'test_central_coveragerc* %s *' % SCRIPT_RESULT,
+        'test_central_coveragerc* %s *' % prop.result,
         '*10 passed*',
     ])
 
@@ -379,25 +400,26 @@ def test_central_coveragerc(testdir):
     assert result.ret == 0
 
 
-def test_show_missing_coveragerc(testdir):
-    script = testdir.makepyfile(SCRIPT)
+def test_show_missing_coveragerc(testdir, prop):
+    script = testdir.makepyfile(prop.code)
     testdir.tmpdir.join('.coveragerc').write("""
 [run]
 source = .
+%s
 
 [report]
 show_missing = true
-""")
+""" % prop.conf)
 
     result = testdir.runpytest('-v',
                                '--cov',
                                '--cov-report=term',
-                               script)
+                               script, *prop.args)
 
     result.stdout.fnmatch_lines([
         '*- coverage: platform *, python * -*',
         'Name * Stmts * Miss * Cover * Missing',
-        'test_show_missing_coveragerc* %s * 11' % SCRIPT_RESULT,
+        'test_show_missing_coveragerc* %s * 11*' % prop.result,
         '*10 passed*',
     ])
 
@@ -479,29 +501,30 @@ def test_foo(foo):
     assert result.ret == 0
 
 
-def test_dist_collocated(testdir):
-    script = testdir.makepyfile(SCRIPT)
-
+def test_dist_collocated(testdir, prop):
+    script = testdir.makepyfile(prop.code)
+    testdir.tmpdir.join('.coveragerc').write(prop.fullconf)
     result = testdir.runpytest('-v',
                                '--cov=%s' % script.dirpath(),
                                '--cov-report=term-missing',
                                '--dist=load',
                                '--tx=2*popen',
                                '--max-slave-restart=0',
-                               script)
+                               script, *prop.args)
 
     result.stdout.fnmatch_lines([
         '*- coverage: platform *, python * -*',
-        'test_dist_collocated* %s *' % SCRIPT_RESULT,
+        'test_dist_collocated* %s *' % prop.result,
         '*10 passed*'
     ])
     assert result.ret == 0
 
 
-def test_dist_not_collocated(testdir):
-    script = testdir.makepyfile(SCRIPT)
+def test_dist_not_collocated(testdir, prop):
+    script = testdir.makepyfile(prop.code)
     dir1 = testdir.mkdir('dir1')
     dir2 = testdir.mkdir('dir2')
+    testdir.tmpdir.join('.coveragerc').write(prop.fullconf)
 
     result = testdir.runpytest('-v',
                                '--cov=%s' % script.dirpath(),
@@ -510,12 +533,13 @@ def test_dist_not_collocated(testdir):
                                '--tx=popen//chdir=%s' % dir1,
                                '--tx=popen//chdir=%s' % dir2,
                                '--rsyncdir=%s' % script.basename,
+                               '--rsyncdir=.coveragerc',
                                '--max-slave-restart=0', '-s',
-                               script)
+                               script, *prop.args)
 
     result.stdout.fnmatch_lines([
         '*- coverage: platform *, python * -*',
-        'test_dist_not_collocated* %s *' % SCRIPT_RESULT,
+        'test_dist_not_collocated* %s *' % prop.result,
         '*10 passed*'
     ])
     assert result.ret == 0
@@ -570,7 +594,6 @@ def test_central_subprocess_change_cwd_with_pythonpath(testdir, monkeypatch):
     stuff.join('child_script.py').write(SCRIPT_CHILD)
     testdir.makefile('', coveragerc="""
 [run]
-branch = true
 parallel = true
 """)
 
@@ -579,6 +602,7 @@ parallel = true
                                '--cov=child_script',
                                '--cov-config=coveragerc',
                                '--cov-report=term-missing',
+                               '--cov-branch',
                                parent_script)
 
     result.stdout.fnmatch_lines([
@@ -597,14 +621,15 @@ def test_foo():
 """)
     testdir.makefile('', coveragerc="""
 [run]
-branch = true
 parallel = true
 omit =
     */__init__.py
 """)
     result = testdir.runpytest('-v',
                                '--cov-config=coveragerc',
-                               '--cov=%s' % script.dirpath(), script)
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-branch',
+                               script)
     result.stdout.fnmatch_lines([
         '*- coverage: platform *, python * -*',
         'test_central_subprocess_no_subscript* * 3 * 0 * 100%*',
@@ -673,14 +698,14 @@ def test_invalid_coverage_source(testdir):
         '*10 passed*'
     ])
     result.stderr.fnmatch_lines([
-        'Coverage.py warning: No data was collected.'
+        'Coverage.py warning: No data was collected.*'
     ])
 
     if StrictVersion(coverage.__version__) <= StrictVersion("3.8"):
         # older `coverage report` doesn't error on missing imports
         assert result.ret == 0
     else:
-        # newer `coverage report` errors on missing importts
+        # newer `coverage report` errors on missing imports
         result.stderr.fnmatch_lines([
             'ERROR: Failed to generate report: No data to report.',
         ])
@@ -752,7 +777,18 @@ def test_funcarg_not_active(testdir):
 def test_multiprocessing_subprocess(testdir):
     py.test.importorskip('multiprocessing.util')
 
-    script = testdir.makepyfile(MULTIPROCESSING_SCRIPT)
+    script = testdir.makepyfile('''
+import multiprocessing
+
+def target_fn():
+    a = True
+    return a
+
+def test_run_target():
+    p = multiprocessing.Process(target=target_fn)
+    p.start()
+    p.join()
+''')
 
     result = testdir.runpytest('-v',
                                '--cov=%s' % script.dirpath(),
@@ -767,6 +803,44 @@ def test_multiprocessing_subprocess(testdir):
     assert result.ret == 0
 
 
+@pytest.mark.skipif('sys.platform == "win32"',
+                    reason="multiprocessing don't support clean process temination on Windows")
+def test_multiprocessing_subprocess_with_terminate(testdir):
+    py.test.importorskip('multiprocessing.util')
+
+    script = testdir.makepyfile('''
+import multiprocessing
+import time
+from pytest_cov.embed import cleanup_on_sigterm
+cleanup_on_sigterm()
+
+event = multiprocessing.Event()
+
+def target_fn():
+    a = True
+    event.set()
+    time.sleep(5)
+
+def test_run_target():
+    p = multiprocessing.Process(target=target_fn)
+    p.start()
+    event.wait(1)
+    p.terminate()
+    p.join()
+''')
+
+    result = testdir.runpytest('-v',
+                               '--cov=%s' % script.dirpath(),
+                               '--cov-report=term-missing',
+                               script)
+
+    result.stdout.fnmatch_lines([
+        '*- coverage: platform *, python * -*',
+        'test_multiprocessing_subprocess* 15 * 100%*',
+        '*1 passed*'
+    ])
+    assert result.ret == 0
+
 MODULE = '''
 def func():
     return 1
@@ -997,7 +1071,10 @@ def test_dist_bare_cov(testdir):
 
 
 def test_not_started_plugin_does_not_fail(testdir):
-    plugin = pytest_cov.plugin.CovPlugin(None, None, start=False)
+    class ns:
+        cov_source = [True]
+        cov_report = ''
+    plugin = pytest_cov.plugin.CovPlugin(ns, None, start=False)
     plugin.pytest_runtestloop(None)
     plugin.pytest_terminal_summary(None)
 
@@ -1084,45 +1161,95 @@ def test_external_data_file_negative(testdir):
 
 
 @xdist
-def test_append_coverage(testdir, opts):
-    script = testdir.makepyfile(test_1=SCRIPT)
-    testdir.tmpdir.join('.coveragerc').write("")
+def xtest_append_coverage(testdir, opts, prop):
+    script = testdir.makepyfile(test_1=prop.code)
+    testdir.tmpdir.join('.coveragerc').write(prop.fullconf)
     result = testdir.runpytest('-v',
                                '--cov=%s' % script.dirpath(),
                                script,
-                               *opts.split())
+                               *opts.split()+prop.args)
     result.stdout.fnmatch_lines([
-        'test_1* %s*' % SCRIPT_RESULT,
+        'test_1* %s*' % prop.result,
     ])
-    script2 = testdir.makepyfile(test_2=SCRIPT2)
+    script2 = testdir.makepyfile(test_2=prop.code2)
     result = testdir.runpytest('-v',
                                '--cov-append',
                                '--cov=%s' % script2.dirpath(),
                                script2,
-                               *opts.split())
+                               *opts.split()+prop.args)
     result.stdout.fnmatch_lines([
-        'test_1* %s*' % SCRIPT_RESULT,
-        'test_2* %s*' % SCRIPT2_RESULT,
+        'test_1* %s*' % prop.result,
+        'test_2* %s*' % prop.result2,
     ])
 
 
 @xdist
-def test_do_not_append_coverage(testdir, opts):
-    script = testdir.makepyfile(test_1=SCRIPT)
+def xtest_do_not_append_coverage(testdir, opts, prop):
+    script = testdir.makepyfile(test_1=prop.code)
     testdir.tmpdir.join('.coveragerc').write("")
     result = testdir.runpytest('-v',
                                '--cov=%s' % script.dirpath(),
                                script,
-                               *opts.split())
+                               *opts.split()+prop.args)
     result.stdout.fnmatch_lines([
-        'test_1* %s*' % SCRIPT_RESULT,
+        'test_1* %s*' % prop.result,
     ])
-    script2 = testdir.makepyfile(test_2=SCRIPT2)
+    script2 = testdir.makepyfile(test_2=prop.code2)
     result = testdir.runpytest('-v',
                                '--cov=%s' % script2.dirpath(),
                                script2,
-                               *opts.split())
+                               *opts.split()+prop.args)
     result.stdout.fnmatch_lines([
         'test_1* 0%',
-        'test_2* %s*' % SCRIPT2_RESULT,
+        'test_2* %s*' % prop.result2,
     ])
+
+
+def test_pth_failure(monkeypatch):
+    with open('src/pytest-cov.pth') as fh:
+        payload = fh.read()
+
+    class SpecificError(Exception):
+        pass
+
+    def bad_init():
+        raise SpecificError()
+
+    buff = StringIO()
+
+    from pytest_cov import embed
+
+    monkeypatch.setattr(embed, 'init', bad_init)
+    monkeypatch.setattr(sys, 'stderr', buff)
+    monkeypatch.setitem(os.environ, 'COV_CORE_SOURCE', 'foobar')
+    exec_(payload)
+    assert buff.getvalue() == '''pytest-cov: Failed to setup subprocess coverage. Environ: {'COV_CORE_SOURCE': 'foobar'} Exception: SpecificError()
+'''
+
+
+def test_double_cov(testdir):
+    script = testdir.makepyfile(SCRIPT_SIMPLE)
+    result = testdir.runpytest('-v',
+                               '--cov', '--cov=%s' % script.dirpath(),
+                               script)
+
+    result.stdout.fnmatch_lines([
+        '*- coverage: platform *, python * -*',
+        'test_double_cov* %s*' % SCRIPT_SIMPLE_RESULT,
+        '*1 passed*'
+    ])
+    assert result.ret == 0
+
+
+def test_double_cov2(testdir):
+    script = testdir.makepyfile(SCRIPT_SIMPLE)
+    result = testdir.runpytest('-v',
+                               '--cov', '--cov',
+                               script)
+
+    result.stdout.fnmatch_lines([
+        '*- coverage: platform *, python * -*',
+        'test_double_cov2* %s*' % SCRIPT_SIMPLE_RESULT,
+        '*1 passed*'
+    ])
+    assert result.ret == 0
